import json
import os
from typing import Any, AsyncGenerator, Generator, Literal, Sequence, cast

from google import genai  # type: ignore
from google.auth.credentials import Credentials
from google.genai.types import (
    Blob,
    Content,
    EmbedContentConfig,
    FileData,
    FunctionCall,
    FunctionDeclaration,
    FunctionResponse,
    GenerateContentConfig,
    GenerateImageConfig,
    Part,
)
from google.genai.types import Tool as GeminiTool

from ..core.client import LLMClient
from ..core.embedding import EmbeddingResults
from ..core.prompt import (
    AudioPrompt,
    ImagePrompt,
    PDFPrompt,
    Prompt,
    SystemPrompt,
    TextPrompt,
    ToolCallPrompt,
    ToolUsePrompt,
    URIPrompt,
    VideoPrompt,
)
from ..core.response import (
    ImageResponse,
    Response,
    ResponseType,
    TextResponse,
    ToolCallResponse,
)
from ..core.tool import Tool
from ..core.usage import Usage
from ..utils import (
    create_parameters,
    generate_dummy_call_id,
    make_batch,
    snake_to_camel,
    utf8_to_bytes,
)
from .gemini_utils import recurse_transform_type_to_upper

GeminiMessage = Content | str


class GoogleClient(LLMClient[Content, str, Part, GeminiTool]):
    """
    Wrapper client for interacting with the Gemini API.

    Parameters
    ----------
    api_key_env_name : str, optional
        Name of the environment variable containing the API key.
    api_type : str, optional
        Type of API to use. Either "aistudio" or "vertexai".
    project_id_env_name : str, optional
        Name of the environment variable containing the project ID.
    location : str, optional
        Location of the API.
    credentials : Credentials, optional
        Google credentials to use.
    """

    def __init__(
        self,
        api_key_env_name: str | None = None,
        api_type: Literal["aistudio", "vertexai"] = "aistudio",
        project_id_env_name: str | None = None,
        location: str | None = None,
        credentials: Credentials | None = None,
    ):
        self.api_key = os.getenv(api_key_env_name) if api_key_env_name else None
        self.project_id = os.getenv(project_id_env_name) if project_id_env_name else None
        self.location = location
        self.credentials = credentials
        self.vertexai = api_type == "vertexai"
        self.client = genai.Client(
            vertexai=self.vertexai,
            api_key=self.api_key,
            project=self.project_id,
            location=self.location,
            credentials=self.credentials,
        )

    def setup_system_instruction(self, system_instruction: SystemPrompt) -> str:
        """
        Setup the system instruction.

        Parameters
        ----------
        system_instruction : SystemPrompt
            System instruction.

        Returns
        ----------
        str
            List of messages with the system instruction.
        """
        return system_instruction.contents

    def generate_text(
        self, messages: list[Content], system_instruction: str | None = None, **kwargs: Any
    ) -> Response:
        """
        Generate text from a list of messages.

        Parameters
        ----------
        messages : list[Content]
            List of messages to generate text from.
        system_instruction : str, optional
            System instruction to include in the messages.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.GenerateContentConfig`.
            For more details, see the document of the `google-genai` package.

        Returns
        ----------
        Response
            Response object containing generated text.
        """
        if system_instruction:
            kwargs["system_instruction"] = system_instruction

        model = cast(str | None, kwargs.pop("model", None))
        name = cast(str | None, kwargs.pop("name", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        generation_config_params = create_parameters(GenerateContentConfig, None, None, **_kwargs)
        generation_config = GenerateContentConfig(**generation_config_params)

        response = self.client.models.generate_content(
            model=model,
            contents=messages,
            config=generation_config,
        )

        contents: list[TextResponse | ToolCallResponse] = []
        for candidate in response.candidates:
            if candidate.content:
                for part in candidate.content.parts:
                    if part.text:
                        if text := part.text.strip():
                            contents.append(TextResponse(text=text))
                    elif part.function_call:
                        call_id = part.function_call.id or generate_dummy_call_id(n=24)
                        contents.append(
                            ToolCallResponse(
                                call_id=call_id,
                                name=part.function_call.name,
                                args=json.dumps(part.function_call.args),
                            )
                        )

        return Response(
            contents=cast(list[ResponseType], contents),
            usage=Usage(
                model_name=model,
                prompt_tokens=response.usage_metadata.prompt_token_count or 0,
                output_tokens=response.usage_metadata.candidates_token_count or 0,
            ),
            raw=response,
            name=name,
        )

    async def generate_text_async(
        self, messages: list[Content], system_instruction: str | None = None, **kwargs: Any
    ) -> Response:
        """
        Generate text from a list of messages asynchronously.

        Parameters
        ----------
        messages : list[Content]
            List of messages to generate text from.
        system_instruction : str, optional
            System instruction.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.GenerateContentConfig`.
            For more details, see the document of the `google-genai` package.

        Returns
        ----------
        Response
            Response object containing generated text.
        """
        if system_instruction:
            kwargs["system_instruction"] = system_instruction

        model = cast(str | None, kwargs.pop("model", None))
        name = cast(str | None, kwargs.pop("name", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        generation_config_params = create_parameters(GenerateContentConfig, None, None, **_kwargs)
        generation_config = GenerateContentConfig(**generation_config_params)

        response = await self.client.aio.models.generate_content(
            model=model,
            contents=messages,
            config=generation_config,
        )

        contents: list[TextResponse | ToolCallResponse] = []
        for candidate in response.candidates:
            if candidate.content:
                for part in candidate.content.parts:
                    if part.text:
                        if text := part.text.strip():
                            contents.append(TextResponse(text=text))
                    elif part.function_call:
                        call_id = part.function_call.id or generate_dummy_call_id(n=24)
                        contents.append(
                            ToolCallResponse(
                                call_id=call_id,
                                name=part.function_call.name,
                                args=json.dumps(part.function_call.args),
                            )
                        )

        return Response(
            contents=cast(list[ResponseType], contents),
            usage=Usage(
                model_name=model,
                prompt_tokens=response.usage_metadata.prompt_token_count or 0,
                output_tokens=response.usage_metadata.candidates_token_count or 0,
            ),
            raw=response,
            name=name,
        )

    def stream_text(
        self, messages: list[Content], system_instruction: str | None = None, **kwargs: Any
    ) -> Generator[Response, None, None]:
        """
        Stream text from a list of messages.

        Parameters
        ----------
        messages : list[Content]
            List of messages to stream text from.
        system_instruction : str, optional
            System instruction.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.GenerateContentConfig`.
            For more details, see the document of the `google-genai` package.

        Yields
        ----------
        Response
            Response object containing combined chunk text or args.
        """
        if system_instruction:
            kwargs["system_instruction"] = system_instruction

        model = cast(str | None, kwargs.pop("model", None))
        name = cast(str | None, kwargs.pop("name", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        generation_config_params = create_parameters(GenerateContentConfig, None, None, **_kwargs)
        generation_config = GenerateContentConfig(**generation_config_params)

        streamed_response = self.client.models.generate_content_stream(
            model=model,
            contents=messages,
            config=generation_config,
        )

        chunk_texts = ""
        usage = Usage(model_name=model)
        res: TextResponse | ToolCallResponse | None = None
        contents: list[TextResponse | ToolCallResponse] = []
        for chunk in streamed_response:
            if usage_metadata := chunk.usage_metadata:
                usage = Usage(
                    model_name=model,
                    prompt_tokens=usage_metadata.prompt_token_count or 0,
                    output_tokens=usage_metadata.candidates_token_count or 0,
                )

            for candidate in chunk.candidates:
                for part in candidate.content.parts:
                    if part.text and part.text.strip():
                        chunk_texts += part.text
                        res = TextResponse(text=chunk_texts)

                        yield Response(
                            contents=[res],
                            usage=usage,
                            raw=chunk,
                            name=name,
                        )
                        continue
                    elif part.text and not part.text.strip():
                        continue

                    if res:
                        contents.append(res)
                        res = None

                    chunk_texts = ""

                    if part.function_call:  # non-streaming
                        call_id = part.function_call.id or generate_dummy_call_id(n=24)
                        args = json.dumps(part.function_call.args)
                        res = ToolCallResponse(
                            call_id=call_id,
                            name=part.function_call.name,
                            args=args,
                        )
                        yield Response(
                            contents=[res],
                            usage=usage,
                            raw=chunk,
                            name=name,
                        )

                        if res:
                            contents.append(res)
                            res = None

        if res:
            contents.append(res)
            res = None

        yield Response(
            contents=cast(list[ResponseType], contents),
            usage=usage,
            raw=chunk,
            name=name,
            is_last_chunk=True,
        )

    async def stream_text_async(
        self, messages: list[Content], system_instruction: str | None = None, **kwargs: Any
    ) -> AsyncGenerator[Response, None]:
        """
        Stream text from a list of messages asynchronously.

        Parameters
        ----------
        messages : list[Content]
            List of messages to stream text from.
        system_instruction : str, optional
            System instruction.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.GenerateContentConfig`.
            For more details, see the document of the `google-genai` package.

        Yields
        ----------
        Response
            Response object containing combined chunk text or args.
        """
        if system_instruction:
            kwargs["system_instruction"] = system_instruction

        model = cast(str | None, kwargs.pop("model", None))
        name = cast(str | None, kwargs.pop("name", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        generation_config_params = create_parameters(GenerateContentConfig, None, None, **_kwargs)
        generation_config = GenerateContentConfig(**generation_config_params)

        streamed_response = self.client.aio.models.generate_content_stream(
            model=model,
            contents=messages,
            config=generation_config,
        )

        chunk_texts = ""
        usage = Usage(model_name=model)
        res: TextResponse | ToolCallResponse | None = None
        contents: list[TextResponse | ToolCallResponse] = []
        async for chunk in streamed_response:
            if usage_metadata := chunk.usage_metadata:
                usage = Usage(
                    model_name=model,
                    prompt_tokens=usage_metadata.prompt_token_count or 0,
                    output_tokens=usage_metadata.candidates_token_count or 0,
                )

            for candidate in chunk.candidates:
                for part in candidate.content.parts:
                    if part.text and part.text.strip():
                        chunk_texts += part.text
                        res = TextResponse(text=chunk_texts)

                        yield Response(
                            contents=[res],
                            usage=usage,
                            raw=chunk,
                            name=name,
                        )
                        continue
                    elif part.text and not part.text.strip():
                        continue

                    if res:
                        contents.append(res)
                        res = None

                    chunk_texts = ""

                    if part.function_call:  # non-streaming
                        call_id = part.function_call.id or generate_dummy_call_id(n=24)
                        args = json.dumps(part.function_call.args)
                        res = ToolCallResponse(
                            call_id=call_id,
                            name=part.function_call.name,
                            args=args,
                        )
                        yield Response(
                            contents=[res],
                            usage=usage,
                            raw=chunk,
                            name=name,
                        )

                        if res:
                            contents.append(res)
                            res = None

        if res:
            contents.append(res)
            res = None

        yield Response(
            contents=cast(list[ResponseType], contents),
            usage=usage,
            raw=chunk,
            name=name,
            is_last_chunk=True,
        )

    def embed_text(self, texts: Sequence[str], **kwargs: Any) -> EmbeddingResults:
        """
        Embed text into a vector space.

        Parameters
        ----------
        texts : Sequence[str]
            Texts to embed.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.EmbedContentConfig`.
            For more details, see the document of the `google-genai` package.

        Returns
        ----------
        EmbeddingResults
            Embedding results.
        """
        if not (isinstance(texts, list) or isinstance(texts, str)):
            raise ValueError("Texts must be a string, not a list of strings.")

        if not isinstance(texts, list):
            texts = [texts]

        model = cast(str | None, kwargs.pop("model", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        embed_config_params = create_parameters(EmbedContentConfig, None, None, **_kwargs)
        embed_config = EmbedContentConfig(**embed_config_params)

        total_usage = Usage(model_name=model)
        batch_size = kwargs.get("batch_size", 10)
        embeddings = []
        for batch in make_batch(texts, batch_size=batch_size):
            embedding = self.client.models.embed_content(
                model=model,
                contents=batch,
                config=embed_config,
            )
            embeddings.extend([emb.values for emb in embedding.embeddings])
            if matadata := embedding.metadata:
                if billable_character_count := matadata.billable_character_count:
                    total_usage += Usage(prompt_tokens=billable_character_count or 0)

        return EmbeddingResults(
            text=texts,
            embeddings=embeddings,
            usage=total_usage,
        )

    async def embed_text_async(self, texts: Sequence[str], **kwargs: Any) -> EmbeddingResults:
        """
        Embed text into a vector space asynchronously.

        Parameters
        ----------
        texts : Sequence[str]
            Texts to embed.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.EmbedContentConfig`.
            For more details, see the document of the `google-genai` package.

        Returns
        ----------
        EmbeddingResults
            Embedding results.
        """
        if not (isinstance(texts, list) or isinstance(texts, str)):
            raise ValueError("Texts must be a string, not a list of strings.")

        if not isinstance(texts, list):
            texts = [texts]

        model = cast(str | None, kwargs.pop("model", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        embed_config_params = create_parameters(EmbedContentConfig, None, None, **_kwargs)
        embed_config = EmbedContentConfig(**embed_config_params)

        total_usage = Usage(model_name=model)
        batch_size = kwargs.get("batch_size", 10)
        embeddings = []
        for batch in make_batch(texts, batch_size=batch_size):
            embedding = await self.client.aio.models.embed_content(
                model=model,
                contents=batch,
                config=embed_config,
            )
            embeddings.extend([emb.values for emb in embedding.embeddings])
            if matadata := embedding.metadata:
                if billable_character_count := matadata.billable_character_count:
                    total_usage += Usage(prompt_tokens=billable_character_count or 0)

        return EmbeddingResults(
            text=texts,
            embeddings=embeddings,
            usage=total_usage,
        )

    def generate_image(self, prompt: str, **kwargs: Any) -> Response:
        """
        Generate an image from a prompt.

        Parameters
        ----------
        prompt : str
            Prompt to generate an image from.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.GenerateImageConfig`.
            For more details, see the document of the `google-genai` package.

        Returns
        ----------
        Response
            Response object containing generated image.
        """

        model = cast(str | None, kwargs.pop("model", None))
        name = cast(str | None, kwargs.pop("name", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        generate_config_params = create_parameters(GenerateImageConfig, None, None, **_kwargs)
        generate_config = GenerateImageConfig(**generate_config_params)

        image = self.client.models.generate_image(
            model=model,
            prompt=prompt,
            config=generate_config,
        )

        contents = []
        for generated_image in image.generated_images:
            pil_image = generated_image.image._pil_image
            img_format = pil_image.format.lower() if pil_image.format else "png"
            contents.append(ImageResponse(image=pil_image, format=img_format))

        return Response(
            contents=contents,
            usage=Usage(),
            raw=image,
            name=name,
        )

    async def generate_image_async(self, prompt: str, **kwargs: Any) -> Response:
        """
        Generate an image from a prompt asynchronously.

        Parameters
        ----------
        prompt : str
            Prompt to generate an image from.
        **kwargs : Any
            Additional keyword arguments.
            Basically the same as the parameters in `google.genai.types.GenerateImageConfig`.
            For more details, see the document of the `google-genai` package.

        Returns
        ----------
        Response
            Response object containing generated image.
        """
        model = cast(str | None, kwargs.pop("model", None))
        name = cast(str | None, kwargs.pop("name", None))

        _kwargs = {snake_to_camel(k): v for k, v in kwargs.items()}
        generate_config_params = create_parameters(GenerateImageConfig, None, None, **_kwargs)
        generate_config = GenerateImageConfig(**generate_config_params)

        image = await self.client.aio.models.generate_image(
            model=model,
            prompt=prompt,
            config=generate_config,
        )

        contents = []
        for generated_image in image.generated_images:
            pil_image = generated_image.image._pil_image
            img_format = pil_image.format.lower() if pil_image.format else "png"
            contents.append(ImageResponse(image=pil_image, format=img_format))

        return Response(
            contents=contents,
            usage=Usage(),
            raw=image,
            name=name,
        )

    def map_to_client_prompts(self, messages: list[Prompt]) -> list[Content]:
        """
        Map a message to a provider-specific representation.

        Parameters
        ----------
        message : list[Prompt]
            List of prompts to map.

        Returns
        ----------
        list[Content]
            List of provider-specific message representation.
        """
        mapped_messages: list[Content] = []
        for message in messages:
            if not message.contents:
                continue

            parts: list[Part] = []
            for content in message.contents:
                if isinstance(content, str):
                    parts.append(Part(text=content))
                elif isinstance(content, TextPrompt):
                    parts.append(Part(text=content.text))
                elif isinstance(content, ImagePrompt):
                    parts.append(
                        Part(
                            inline_data=Blob(
                                data=utf8_to_bytes(cast(str, content.image)),
                                mime_type=f"image/{content.format}",
                            )
                        )
                    )
                elif isinstance(content, PDFPrompt):
                    parts.extend(
                        [
                            Part(
                                inline_data=Blob(
                                    data=utf8_to_bytes(cast(str, img.image)),
                                    mime_type="image/jpeg",
                                )
                            )
                            for img in content.as_image_content()
                        ]
                    )
                elif isinstance(content, URIPrompt):
                    parts.append(
                        Part(file_data=FileData(file_uri=content.uri, mime_type=content.mime_type))
                    )
                elif isinstance(content, VideoPrompt):
                    parts.extend(
                        [
                            Part(
                                inline_data=Blob(
                                    data=utf8_to_bytes(cast(str, img.image)),
                                    mime_type="image/jpeg",
                                )
                            )
                            for img in content.as_image_content()
                        ]
                    )
                elif isinstance(content, AudioPrompt):
                    parts.append(
                        Part(
                            inline_data=Blob(
                                data=utf8_to_bytes(cast(str, content.audio)),
                                mime_type=content.mime_type,
                            )
                        )
                    )
                elif isinstance(content, ToolCallPrompt):
                    content_args = content.args
                    if isinstance(content_args, str) and content_args:
                        content_args = json.loads(content_args)

                    parts.append(
                        Part(
                            function_call=FunctionCall(
                                id=content.call_id
                                if not self.vertexai
                                else None,  # VertexAI does not support id
                                args=content_args,
                                name=content.name,
                            )
                        )
                    )
                elif isinstance(content, ToolUsePrompt):
                    if content.output:
                        parts.append(
                            Part(
                                function_response=FunctionResponse(
                                    id=content.call_id
                                    if not self.vertexai
                                    else None,  # VertexAI does not support id
                                    name=content.name,
                                    response={"output": content.output},
                                )
                            )
                        )
                    elif content.error:
                        parts.append(
                            Part(
                                function_response=FunctionResponse(
                                    id=content.call_id
                                    if not self.vertexai
                                    else None,  # VertexAI does not support id
                                    name=content.name,
                                    response={"error": content.error},
                                )
                            )
                        )

            mapped_messages.append(Content(role=message.role, parts=parts))

        return mapped_messages

    def map_to_client_tools(self, tools: list[Tool], **kwargs: Any) -> list[GeminiTool]:
        """
        Map tools to provider-specific representations.

        Parameters
        ----------
        tools : list[Tool]
            List of tools to map.

        Returns
        ----------
        list[GeminiTool]
            List of provider-specific tool representations.
        """

        return [
            GeminiTool(
                function_declarations=[self.map_to_client_tool(tool=tool) for tool in tools],
                **kwargs,
            )
        ]

    def map_to_client_tool(self, tool: Tool, **kwargs: Any) -> FunctionDeclaration:
        if tool.schema_dict is None:
            raise ValueError("Tool schema is required.")

        schema = recurse_transform_type_to_upper(tool.schema_dict)
        return FunctionDeclaration(
            description=tool.description,
            parameters=schema
            if schema.get("properties")
            else None,  # Google Developer API does not support empty properties
            response=kwargs.get("response"),
            name=tool.name,
        )



from .core.agent import Agent
from .core.config import AgentConfig
from .core.internal_prompt import InternalPrompt
from .core.logger import get_logger
from .core.model import LLMModel
from .core.prompt import (
    AudioPrompt,
    ImagePrompt,
    PDFPrompt,
    Prompt,
    PromptType,
    SystemPrompt,
    TextPrompt,
    ToolCallPrompt,
    ToolUsePrompt,
    URIPrompt,
    VideoPrompt,
)
from .core.pydantic import BaseModel
from .core.response import (
    AudioResponse,
    ImageResponse,
    Response,
    ResponseType,
    TextResponse,
    ToolCallResponse,
    VideoResponse,
)
from .core.tool import Tool
from .core.usage import Usage
from .memory.in_memory import InMemoryConversationMemory
from .memory.json import JSONConversationMemory
from .memory.pickle import PickleConversationMemory
from .prompt_template import PromptTemplate



import base64
import io
from pathlib import Path
from typing import Any, Literal, cast

import numpy as np
from PIL import Image
from pydantic import model_validator

from ..utils import decode_image, encode_image
from .pydantic import BaseModel
from .typing import ArrayType, PathType
from .usage import NamedUsage, Usage


class TextResponse(BaseModel):
    text: str


class ToolCallResponse(BaseModel):
    name: str | None = None
    args: str | None = None
    call_id: str | None = None


class ImageResponse(BaseModel):
    image: str | Image.Image
    format: str | None = None

    @model_validator(mode="before")
    @classmethod
    def check_image_type(cls: "ImageResponse", data: Any) -> Any:  # type: ignore[misc]
        if isinstance(data, dict) and isinstance(data.get("image"), Image.Image):
            data["image"] = encode_image(image=cast(Image.Image, data.get("image")))

        return data

    def decode(self) -> Image.Image:
        return decode_image(cast(str, self.image))

    def __str__(self) -> str:
        if isinstance(self.image, str):
            return f"ImageResponse(image={self.image[:20]}..., format={self.format})"
        else:
            raise ValueError("Invalid image type")

    def __repr__(self) -> str:
        return self.__str__()


class AudioResponse(BaseModel):
    audio: PathType | bytes | ArrayType | str
    sr: int | None = None
    mime_type: str | None = None
    audio_id: str | None = None

    def asarray(self) -> ArrayType:
        import soundfile as sf

        if isinstance(self.audio, np.ndarray):
            return self.audio
        elif isinstance(self.audio, bytes):
            return cast(ArrayType, sf.read(io.BytesIO(self.audio))[0])
        elif isinstance(self.audio, str):
            return cast(ArrayType, sf.read(io.BytesIO(base64.b64decode(self.audio)))[0])
        elif isinstance(self.audio, Path):
            return cast(ArrayType, sf.read(self.audio)[0])
        else:
            raise ValueError("Invalid audio type")

    def asbytes(self) -> bytes:
        if isinstance(self.audio, bytes):
            return self.audio
        elif isinstance(self.audio, str):
            return base64.b64decode(self.audio)
        else:
            raise ValueError("Invalid audio type")

    @staticmethod
    def to_bytes(audio: ArrayType, sr: int, mime_type: str) -> bytes:
        import soundfile as sf

        buffer = io.BytesIO()
        sf.write(
            file=buffer,
            data=audio,
            samplerate=sr,
            format=mime_type.split("/")[-1].upper() if mime_type else None,
        )

        buffer.seek(0)
        return buffer.read()

    @model_validator(mode="before")
    @classmethod
    def setup(cls: "AudioResponse", data: Any) -> Any:  # type: ignore[misc]
        try:
            import soundfile as sf
        except ModuleNotFoundError:
            return data

        if isinstance(data, dict) and "audio" in data:
            try:
                assert Path(data["audio"]).is_file()
                file_format = Path(data["audio"]).suffix.lstrip(".").lower()
                if file_format in ["wav", "mp3", "aiff", "ogg", "flac"]:
                    data["mime_type"] = f"audio/{file_format}"
                elif file_format in ["mp4", "mpeg", "mov", "avi", "wmv", "mpg"]:
                    data["audio"], data["sr"] = cls._from_video(data["audio"])
                    data["mime_type"] = "audio/wav"
                else:
                    raise ValueError(f"Invalid audio file format: {file_format}")
            except OSError:
                data["mime_type"] = "audio/wav"
            else:
                pass

            if isinstance(data["audio"], (str, Path)):
                try:
                    assert Path(data["audio"]).is_file()
                    if isinstance(data["audio"], Path):
                        data["audio"] = data["audio"].as_posix()
                    data["audio"], data["sr"] = sf.read(data["audio"])
                except OSError:
                    data["audio"], data["sr"] = sf.read(io.BytesIO(base64.b64decode(data["audio"])))
                else:
                    pass

            if isinstance(data["audio"], bytes):
                data["audio"], data["sr"] = sf.read(io.BytesIO(data["audio"]))
            elif isinstance(data["audio"], io.BytesIO):
                data["audio"], data["sr"] = sf.read(data["audio"])
            elif isinstance(data["audio"], np.ndarray):
                pass
            elif isinstance(data["audio"], list):
                data["audio"] = np.array(data["audio"])
            else:
                raise ValueError("Invalid audio data type")

            if isinstance(data["audio"], np.ndarray):
                assert data.get("sr") is not None, "Sample rate must be provided"
                data["audio"] = cls._convert_stereo_to_mono(cast(ArrayType, data["audio"]))
                audio_bytes = cls.to_bytes(
                    data["audio"],
                    data["sr"],
                    data["mime_type"],
                )
                data["audio"] = base64.b64encode(audio_bytes).decode("utf-8")
        return data

    def __str__(self) -> str:
        if isinstance(self.audio, str):
            return f"AudioResponse(audio={self.audio[:20]}..., sr={self.sr}, mime_type={self.mime_type})"
        else:
            raise ValueError("Invalid audio type")

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def _convert_stereo_to_mono(data: ArrayType) -> ArrayType:
        if data.ndim == 2:
            return cast(ArrayType, np.mean(data, axis=1))
        return data

    @staticmethod
    def _from_video(video_file: PathType) -> tuple[ArrayType, int]:
        from moviepy.editor import VideoFileClip

        clip = VideoFileClip(Path(video_file).as_posix())
        audio = clip.audio
        fps = audio.fps
        audio_array = np.array(list(audio.iter_frames()))
        return audio_array, fps


class VideoResponse(BaseModel):
    video: str


class EmbeddingResponse(BaseModel):
    embeddings: list[list[float]]


ResponseType = str | TextResponse | ImageResponse | AudioResponse | VideoResponse | ToolCallResponse

Responses = (
    TextResponse,
    ImageResponse,
    AudioResponse,
    VideoResponse,
    ToolCallResponse,
)


class Response(BaseModel):
    type: Literal["Response"] = "Response"
    role: Literal["assistant"] = "assistant"
    contents: list[ResponseType]
    usage: Usage | NamedUsage | None = None
    raw: Any | None = None
    name: str | None = None
    is_last_chunk: bool | None = None
    prompt: Any | None = None

    @model_validator(mode="before")
    @classmethod
    def setup(cls: "Response", data: Any) -> Any:  # type: ignore[misc]
        if isinstance(data, dict) and "contents" in data:
            if not isinstance(data["contents"], list):
                data["contents"] = [data["contents"]]

            data["contents"] = cls.to_universal_contents(data["contents"])

            if data.get("usage") is None:
                data["usage"] = Usage()

        return data

    @staticmethod
    def _to_universal_contents(
        contents: list[ResponseType] | list[dict[str, Any]] | list[Image.Image],
    ) -> list[ResponseType]:
        _contents: list[ResponseType] = []

        for content in contents:
            if isinstance(content, str):
                content = TextResponse(text=content)
            elif isinstance(content, Image.Image):
                content = ImageResponse(image=content)

            if isinstance(
                content,
                (
                    TextResponse,
                    ImageResponse,
                    ToolCallResponse,
                    VideoResponse,
                    AudioResponse,
                ),
            ):
                _contents.append(content)

            elif isinstance(content, dict):
                for response in Responses:
                    try:
                        _contents.append(response(**content))
                        break
                    except Exception:
                        pass
            else:
                raise NotImplementedError(f"Response type {type(content)} not implemented")

        return _contents

    @staticmethod
    def to_universal_contents(
        contents: list[ResponseType] | Image.Image,
    ) -> list[ResponseType]:
        if isinstance(contents, (ResponseType, list)):
            return Response._to_universal_contents(contents=contents)
        elif isinstance(contents, Image.Image):
            return [ImageResponse(image=contents)]
        else:
            raise ValueError("Invalid contents type.")



from typing import Any

from .pydantic import BaseModel
from .usage import Usage


class RetrievalResults(BaseModel):
    ids: list[int | str]
    documents: list[str]
    metadatas: list[dict[str, Any]] | list[None] | None
    scores: list[float]
    collections: list[str]
    usage: Usage = Usage(prompt_tokens=0, completion_tokens=0)



import json
import os
from typing import Any, AsyncGenerator, Generator, Literal, cast

from anthropic import (
    Anthropic,
    AnthropicBedrock,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    AsyncAnthropicVertex,
)
from anthropic._types import NOT_GIVEN, NotGiven
from anthropic.types import (
    ContentBlock,
    ImageBlockParam,
    InputJSONDelta,
    Message,
    MessageParam,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
    TextBlock,
    TextBlockParam,
    TextDelta,
    ToolParam,
    ToolResultBlockParam,
    ToolUseBlock,
    ToolUseBlockParam,
)
from anthropic.types.image_block_param import Source

from ..core.client import LLMClient
from ..core.prompt import (
    AudioPrompt,
    ImagePrompt,
    PDFPrompt,
    Prompt,
    SystemPrompt,
    TextPrompt,
    ToolCallPrompt,
    ToolUsePrompt,
    URIPrompt,
    VideoPrompt,
)
from ..core.response import Response, TextResponse, ToolCallResponse
from ..core.tool import Tool
from ..core.usage import Usage
from ..utils import create_parameters

ClientType = (
    Anthropic
    | AnthropicBedrock
    | AnthropicVertex
    | AsyncAnthropic
    | AsyncAnthropicBedrock
    | AsyncAnthropicVertex
)

AnthropicContentType = (
    TextBlockParam | ImageBlockParam | ToolUseBlockParam | ToolResultBlockParam | ContentBlock
)


class AnthropicClient(LLMClient[MessageParam, str, AnthropicContentType, ToolParam]):
    """
    Wrapper client for the Anthropic API.

    Parameters
    ----------
    api_key_env_name : str, optional
        Environment variable name for the API key, by default None
    api_type : Literal["anthropic", "bedrock", "vertexai"], optional
        API type, by default "anthropic"
    aws_access_key_env_name : str, optional
        Environment variable name for the AWS access key, by default None
    aws_secret_key_env_name : str, optional
        Environment variable name for the AWS secret key, by default None
    aws_region : str, optional
        AWS region, by default None
    google_cloud_project_env_name : str, optional
        Environment variable name for the Google Cloud project, by default NOT_GIVEN
    google_cloud_region : str, optional
        Google Cloud region, by default NOT_GIVEN
    **kwargs : Any
        Additional parameters to pass to the client.
        Basically, any parameter that the client accepts can be passed here.
        For more details, see the documentation of the anthropic api.
    """

    def __init__(
        self,
        api_key_env_name: str | None = None,
        api_type: Literal["anthropic", "bedrock", "vertexai"] = "anthropic",
        aws_access_key_env_name: str | None = None,
        aws_secret_key_env_name: str | None = None,
        aws_region: str | None = None,
        google_cloud_project_env_name: str | NotGiven = NOT_GIVEN,
        google_cloud_region: str | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ):
        self.api_key = os.getenv(api_key_env_name) if api_key_env_name else None
        self.api_type = api_type
        self.aws_access_key = (
            os.getenv(aws_access_key_env_name) if aws_access_key_env_name else None
        )
        self.aws_secret_key = (
            os.getenv(aws_secret_key_env_name) if aws_secret_key_env_name else None
        )
        self.aws_region = aws_region
        self.google_cloud_project = (
            os.getenv(google_cloud_project_env_name) if google_cloud_project_env_name else NOT_GIVEN
        )
        self.google_cloud_region = google_cloud_region

        self._client: ClientType
        self._async_client: ClientType
        if self.api_type == "anthropic":
            self._client = Anthropic(
                api_key=self.api_key,
                **kwargs,
            )
            self._async_client = AsyncAnthropic(
                api_key=self.api_key,
                **kwargs,
            )
        elif self.api_type == "bedrock":
            self._client = AnthropicBedrock(
                aws_access_key=self.aws_access_key,
                aws_secret_key=self.aws_secret_key,
                aws_region=self.aws_region,
                **kwargs,
            )
            self._async_client = AsyncAnthropicBedrock(
                aws_access_key=self.aws_access_key,
                aws_secret_key=self.aws_secret_key,
                aws_region=self.aws_region,
                **kwargs,
            )
        elif self.api_type == "vertexai":
            self._client = AnthropicVertex(
                region=self.google_cloud_region,
                project_id=self.google_cloud_project,
                **kwargs,
            )
            self._async_client = AsyncAnthropicVertex(
                region=self.google_cloud_region,
                project_id=self.google_cloud_project,
                **kwargs,
            )

    def setup_system_instruction(self, system_instruction: SystemPrompt) -> str:
        """
        Setup the system instruction.

        Parameters
        ----------
        system_instruction : SystemPrompt
            System instruction.

        Returns
        ----------
        str
            List of messages with the system instruction.
        """
        return system_instruction.contents

    def generate_text(
        self, messages: list[MessageParam], system_instruction: str | None = None, **kwargs: Any
    ) -> Response:
        """
        Generate text from a list of messages.

        Parameters
        ----------
        messages : list[MessageParam]
            List of messages to generate text from.
        system_instruction : str, optional
            System instruction, by default None.
        **kwargs : Any
            Additional parameters to pass to the client.
            Same as the parameters accepted by the client.
            For more details, see the documentation of the anthropic api.

        Returns
        ----------
        Response
            Generated text response.
        """
        name = cast(str | None, kwargs.pop("name", None))

        if system_instruction:
            kwargs["system"] = system_instruction

        assert not kwargs.get("stream"), "Use stream_text or stream_text_async for streaming."
        completion_params = create_parameters(self._client.messages.create, **kwargs)
        response: Message = self._client.messages.create(
            messages=messages, stream=False, **completion_params
        )

        contents: list[TextResponse | ToolCallResponse] = []
        for content in response.content:
            if isinstance(content, TextBlock):
                if content.text:
                    if text := content.text.strip():
                        contents.append(TextResponse(text=text))
            elif isinstance(content, ToolUseBlock):
                contents.append(
                    ToolCallResponse(
                        call_id=content.id, name=content.name, args=json.dumps(content.input)
                    )
                )

        return Response(
            contents=contents,
            usage=Usage(
                model_name=response.model,
                prompt_tokens=response.usage.input_tokens,
                output_tokens=response.usage.output_tokens,
            ),
            raw=response,
            name=name,
            prompt=messages,
        )

    async def generate_text_async(
        self, messages: list[MessageParam], system_instruction: str | None = None, **kwargs: Any
    ) -> Response:
        """
        Generate text from a list of messages asynchronously.

        Parameters
        ----------
        messages : list[MessageParam]
            List of messages to generate text from.
        system_instruction : str, optional
            System instruction, by default None.
        **kwargs : Any
            Additional parameters to pass to the client.
            Same as the parameters accepted by the client.
            For more details, see the documentation of the anthropic api.

        Returns
        ----------
        Response
            Generated text response.
        """
        name = cast(str | None, kwargs.pop("name", None))

        if system_instruction:
            kwargs["system"] = system_instruction

        assert not kwargs.get("stream"), "Use stream_text or stream_text_async for streaming."
        completion_params = create_parameters(self._client.messages.create, **kwargs)
        response: Message = await self._async_client.messages.create(
            messages=messages, stream=False, **completion_params
        )

        contents: list[TextResponse | ToolCallResponse] = []
        for content in response.content:
            if isinstance(content, TextBlock):
                if content.text:
                    if text := content.text.strip():
                        contents.append(TextResponse(text=text))
            elif isinstance(content, ToolUseBlock):
                contents.append(
                    ToolCallResponse(
                        call_id=content.id, name=content.name, args=json.dumps(content.input)
                    )
                )

        return Response(
            contents=contents,
            usage=Usage(
                model_name=response.model,
                prompt_tokens=response.usage.input_tokens,
                output_tokens=response.usage.output_tokens,
            ),
            raw=response,
            name=name,
            prompt=messages,
        )

    def stream_text(
        self, messages: list[MessageParam], system_instruction: str | None = None, **kwargs: Any
    ) -> Generator[Response, None, None]:
        """
        Stream text from a list of messages.

        Parameters
        ----------
        messages : list[MessageParam]
            List of messages to stream text from.
        system_instruction : str, optional
            System instruction, by default None.
        **kwargs : Any
            Additional parameters to pass to the client.
            Same as the parameters accepted by the client.
            For more details, see the documentation of the anthropic api.

        Yields
        ----------
        Response
            Streamed text response.
        """
        name = cast(str | None, kwargs.pop("name", None))

        if system_instruction:
            kwargs["system"] = system_instruction

        completion_params = create_parameters(self._client.messages.create, **kwargs)
        streamed_response: Message = self._client.messages.create(
            messages=messages, stream=True, **completion_params
        )

        chunk_texts = ""
        chunk_args = ""
        res: TextResponse | ToolCallResponse | None = None
        contents: list[TextResponse | ToolCallResponse] = []
        with streamed_response as stream:
            for response in stream:
                if isinstance(response, RawMessageStartEvent):
                    usage = Usage(
                        model_name=kwargs.get("model"),
                        prompt_tokens=response.message.usage.input_tokens,
                        output_tokens=response.message.usage.output_tokens,
                    )
                elif isinstance(response, RawContentBlockStartEvent):
                    if isinstance(response.content_block, TextBlock):
                        chunk_texts += response.content_block.text
                        if chunk_texts and response.content_block.text:
                            res = TextResponse(text=chunk_texts.strip())
                            yield Response(
                                contents=[res],
                                usage=usage,
                            )
                        continue

                    elif isinstance(response.content_block, ToolUseBlock):
                        funcname = response.content_block.name
                        call_id = response.content_block.id
                        chunk_args = ""
                        res = ToolCallResponse(
                            name=funcname, call_id=call_id, args=chunk_args.strip()
                        )
                        # yield Response(contents=[res], usage=usage)
                        continue

                elif isinstance(response, RawContentBlockDeltaEvent):
                    if isinstance(response.delta, TextDelta):
                        chunk_texts += response.delta.text
                        if chunk_texts and response.delta.text:
                            res = TextResponse(text=chunk_texts.strip())
                            yield Response(
                                contents=[res],
                                usage=usage,
                            )
                        continue

                    elif isinstance(response.delta, InputJSONDelta):
                        chunk_args += response.delta.partial_json

                        if chunk_args and response.delta.partial_json:
                            res = ToolCallResponse(
                                name=funcname, call_id=call_id, args=chunk_args.strip()
                            )
                            yield Response(
                                contents=[res],
                                usage=usage,
                            )
                        continue
                elif isinstance(response, RawContentBlockStopEvent):
                    if res:
                        contents.append(res)
                        res = None
                elif isinstance(response, RawMessageStopEvent):
                    pass
                elif isinstance(response, RawMessageDeltaEvent):
                    usage = Usage(
                        model_name=kwargs.get("model"),
                        prompt_tokens=usage.prompt_tokens,
                        output_tokens=usage.output_tokens + response.usage.output_tokens,
                    )

        yield Response(contents=contents, usage=usage, is_last_chunk=True, name=name)

    async def stream_text_async(
        self, messages: list[MessageParam], system_instruction: str | None = None, **kwargs: Any
    ) -> AsyncGenerator[Response, None]:
        """
        Stream text from a list of messages asynchronously.

        Parameters
        ----------
        messages : list[MessageParam]
            List of messages to stream text from.
        system_instruction : str, optional
            System instruction, by default None.
        **kwargs : Any
            Additional parameters to pass to the client.
            Same as the parameters accepted by the client.
            For more details, see the documentation of the anthropic api.

        Yields
        ----------
        Response
            Streamed text response.
        """
        name = cast(str | None, kwargs.pop("name", None))

        if system_instruction:
            kwargs["system"] = system_instruction

        completion_params = create_parameters(self._client.messages.create, **kwargs)
        streamed_response: Message = await self._async_client.messages.create(
            messages=messages, stream=True, **completion_params
        )

        chunk_texts = ""
        chunk_args = ""
        res: TextResponse | ToolCallResponse | None = None
        contents: list[TextResponse | ToolCallResponse] = []
        async with streamed_response as stream:
            async for response in stream:
                if isinstance(response, RawMessageStartEvent):
                    usage = Usage(
                        model_name=kwargs.get("model"),
                        prompt_tokens=response.message.usage.input_tokens,
                        output_tokens=response.message.usage.output_tokens,
                    )
                elif isinstance(response, RawContentBlockStartEvent):
                    if isinstance(response.content_block, TextBlock):
                        chunk_texts += response.content_block.text
                        if chunk_texts and response.content_block.text:
                            res = TextResponse(text=chunk_texts.strip())
                            yield Response(
                                contents=[res],
                                usage=usage,
                            )
                        continue

                    elif isinstance(response.content_block, ToolUseBlock):
                        funcname = response.content_block.name
                        call_id = response.content_block.id
                        chunk_args = ""
                        res = ToolCallResponse(
                            name=funcname, call_id=call_id, args=chunk_args.strip()
                        )
                        # yield Response(contents=[res], usage=usage)
                        continue

                elif isinstance(response, RawContentBlockDeltaEvent):
                    if isinstance(response.delta, TextDelta):
                        chunk_texts += response.delta.text
                        if chunk_texts and response.delta.text:
                            res = TextResponse(text=chunk_texts.strip())
                            yield Response(
                                contents=[res],
                                usage=usage,
                            )
                        continue

                    elif isinstance(response.delta, InputJSONDelta):
                        chunk_args += response.delta.partial_json

                        if chunk_args and response.delta.partial_json:
                            res = ToolCallResponse(
                                name=funcname, call_id=call_id, args=chunk_args.strip()
                            )
                            yield Response(
                                contents=[res],
                                usage=usage,
                            )
                        continue
                elif isinstance(response, RawContentBlockStopEvent):
                    if res:
                        contents.append(res)
                        res = None
                elif isinstance(response, RawMessageStopEvent):
                    pass
                elif isinstance(response, RawMessageDeltaEvent):
                    usage = Usage(
                        model_name=kwargs.get("model"),
                        prompt_tokens=usage.prompt_tokens,
                        output_tokens=usage.output_tokens + response.usage.output_tokens,
                    )

        yield Response(contents=contents, usage=usage, is_last_chunk=True, name=name)

    def map_to_client_prompts(self, messages: list[Prompt]) -> list[MessageParam]:
        """
        Map a message to a provider-specific representation.

        Parameters
        ----------
        message : list[Prompt]
            List of prompts to map.

        Returns
        ----------
        list[MessageParam]
            List of provider-specific message representation.
        """
        mapped_messages: list[MessageParam] = []
        for message in messages:
            if not message.contents:
                continue

            contents: list[AnthropicContentType] = []
            for content in message.contents:
                if isinstance(content, str):
                    contents.append(TextBlockParam(text=content, type="text"))
                elif isinstance(content, TextPrompt):
                    contents.append(TextBlockParam(text=content.text, type="text"))
                elif isinstance(content, ToolCallPrompt):
                    contents.append(
                        ToolUseBlockParam(
                            id="toolu_" + content.call_id.split("_")[-1],
                            name=content.name,
                            input=json.loads(content.args),
                            type="tool_use",
                        )
                    )
                elif isinstance(content, ToolUsePrompt):
                    contents.append(
                        ToolResultBlockParam(
                            tool_use_id="toolu_" + content.call_id.split("_")[-1],
                            type="tool_result",
                            content=[
                                TextBlockParam(
                                    text=content.output if content.output else content.error,
                                    type="text",
                                )
                            ],
                            is_error=bool(content.error),
                        )
                    )
                elif isinstance(content, ImagePrompt):
                    contents.append(
                        ImageBlockParam(
                            source=Source(
                                data=content.image,
                                media_type=f"image/{content.format}",
                                type="base64",
                            ),
                            type="image",
                        ),
                    )
                elif isinstance(content, VideoPrompt):
                    contents.extend(
                        [
                            ImageBlockParam(
                                source=Source(
                                    data=frame.image,
                                    media_type=f"image/{frame.format}",
                                    type="base64",
                                ),
                                type="image",
                            )
                            for frame in content.as_image_content()
                        ]
                    )
                elif isinstance(content, AudioPrompt):
                    raise NotImplementedError
                elif isinstance(content, PDFPrompt):
                    contents.extend(
                        [
                            ImageBlockParam(
                                source=Source(
                                    data=page.image,
                                    media_type=f"image/{page.format}",
                                    type="base64",
                                ),
                                type="image",
                            )
                            for page in content.as_image_content()
                        ]
                    )
                elif isinstance(content, URIPrompt):
                    raise NotImplementedError

            mapped_messages.append(
                MessageParam(
                    role=message.role,
                    content=contents,
                )
            )

        return mapped_messages

    def map_to_client_tools(self, tools: list[Tool]) -> list[ToolParam]:
        """
        Map tools to provider-specific representations.

        Parameters
        ----------
        tools : list[Tool]
            List of tools to map.

        Returns
        ----------
        list[ToolParam]
            List of provider-specific tool representations.
        """
        return [self.map_to_client_tool(tool=tool) for tool in tools]

    def map_to_client_tool(self, tool: Tool, **kwargs: Any) -> ToolParam:
        if tool.schema_dict is None:
            raise ValueError("Tool schema is required.")

        schema = tool.schema_dict
        return ToolParam(
            description=tool.description,
            input_schema=schema,
            name=tool.name,
        )



import asyncio
import inspect
import math
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional

from tqdm import tqdm

from ..core.embedding import EmbeddingResults
from ..core.logger import DEFAULT_LOGGER as default_logger
from ..core.model import LLMModel
from ..core.retrieval import RetrievalResults
from ..utils import make_batch


class AbstractLocalCollectionModule(ABC):
    """
    Base class for collection module.
    """

    @abstractmethod
    def get_client(self) -> Any:
        """
        return the client object
        """
        raise NotImplementedError

    @abstractmethod
    def _create_collection(self, client: Any, collection_name: str) -> None:
        """
        create a collection
        """
        raise NotImplementedError

    @abstractmethod
    def _exists(self, client: Any, collection_name: str) -> bool:
        """
        check if the collection exists
        """
        raise NotImplementedError

    @abstractmethod
    def _upsert(
        self,
        client: Any,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs: Any,
    ) -> None:
        """
        upsert embeddings, documents and metadatas.
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_collection(self, client: Any, collection_name: str) -> None:
        """
        delete the collection
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_record(self, client: Any, collection_name: str, **kwargs: Any) -> None:
        """
        delete the record from the collection. kwargs depends on the clieint.
        """
        raise NotImplementedError

    @abstractmethod
    def as_retriever(
        self,
        n_results: int,
        score_threshold: float,
    ) -> "BaseLocalRetrievalModule":
        """
        return the retrieval module
        """
        raise NotImplementedError


class AbstractRemoteCollectionModule(ABC):
    """
    Base class for collection module.
    """

    @abstractmethod
    def get_client(self) -> Any:
        """
        return the client object
        """
        raise NotImplementedError

    @abstractmethod
    def get_async_client(self) -> Any:
        """
        return the async client object
        """
        raise NotImplementedError

    @abstractmethod
    def _create_collection(self, client: Any, collection_name: str) -> None:
        """
        create a collection
        """
        raise NotImplementedError

    @abstractmethod
    async def _create_collection_async(self, client: Any, collection_name: str) -> None:
        """
        create a collection
        """
        raise NotImplementedError

    @abstractmethod
    def _exists(self, client: Any, collection_name: str) -> bool:
        """
        check if the collection exists
        """
        raise NotImplementedError

    @abstractmethod
    async def _exists_async(self, client: Any, collection_name: str) -> bool:
        """
        check if the collection exists
        """
        raise NotImplementedError

    @abstractmethod
    def _upsert(
        self,
        client: Any,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs: Any,
    ) -> None:
        """
        upsert embeddings, documents and metadatas.
        """
        raise NotImplementedError

    @abstractmethod
    async def _upsert_async(
        self,
        client: Any,
        collection_name: str,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs: Any,
    ) -> None:
        """
        upsert embeddings, documents and metadatas.
        """
        raise NotImplementedError

    @abstractmethod
    async def _delete_collection_async(self, client: Any, collection_name: str) -> None:
        """
        delete the collection
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_collection(self, client: Any, collection_name: str) -> None:
        """
        delete the collection
        """
        raise NotImplementedError

    @abstractmethod
    async def _delete_record_async(self, client: Any, collection_name: str, **kwargs: Any) -> None:
        """
        delete the record from the collection. kwargs depends on the clieint.
        """
        raise NotImplementedError

    @abstractmethod
    def _delete_record(self, client: Any, collection_name: str, **kwargs: Any) -> None:
        """
        delete the record from the collection. kwargs depends on the clieint.
        """
        raise NotImplementedError

    @abstractmethod
    def as_retriever(
        self,
        n_results: int,
        score_threshold: float,
    ) -> "BaseRemoteRetrievalModule":
        """
        return the retrieval module
        """
        raise NotImplementedError


class BaseLocalCollectionModule(AbstractLocalCollectionModule):
    def __init__(
        self,
        persistence_directory: Path | str,
        collection_name: str,
        embedder: LLMModel | None = None,  # type: ignore
        logger: Any | None = None,
        batch_size: int = 100,
    ):
        self.persistence_directory = Path(persistence_directory)
        self.embedder = embedder
        self.collection_name = collection_name
        self.logger = logger or default_logger
        self.batch_size = batch_size

    def create_collection(self) -> None:
        client = self.get_client()
        self._create_collection(client=client, collection_name=self.collection_name)

    def exists(self) -> bool:
        client = self.get_client()
        return self._exists(client=client, collection_name=self.collection_name)

    def upsert(
        self,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs: Any,
    ) -> None:
        self._verify_metadata(metadatas)
        client = self.get_client()
        self._upsert(
            client=client,
            collection_name=self.collection_name,
            ids=ids,
            documents=documents,
            embeddings=embeddings,
            metadatas=metadatas,
            **kwargs,
        )

    def delete_collection(self) -> None:
        client = self.get_client()

        if self._exists(client=client, collection_name=self.collection_name):
            self._delete_collection(client=client, collection_name=self.collection_name)

    def delete_record(self, **kwargs: Any) -> None:
        client = self.get_client()
        self._delete_record(client=client, collection_name=self.collection_name, **kwargs)

    def run(
        self,
        documents: list[str],
        metadatas: Optional[list[dict[str, str]]] = None,
        **kwargs: Any,
    ) -> None:
        client = self.get_client()

        if metadatas is not None and len(documents) != len(metadatas):
            raise ValueError(
                "The length of documents and metadatas must be the same. "
                f"Got {len(documents)} documents and {len(metadatas)} metadatas."
            )

        if metadatas is None:
            metadatas = [{} for _ in range(len(documents))]
        else:
            # check if the key 'collection' or 'document' is included in metadatas
            self._verify_metadata(metadatas)

        ids = [i for i in range(len(documents))]

        documents_batch = make_batch(documents, batch_size=self.batch_size)
        metadatas_batch = make_batch(metadatas, batch_size=self.batch_size)
        ids_batch = make_batch(ids, batch_size=self.batch_size)

        n_batches: int = math.ceil(len(documents) / self.batch_size)

        if not self._exists(client=client, collection_name=self.collection_name):
            self._create_collection(client=client, collection_name=self.collection_name)
            self.logger.info(f"Create collection {self.collection_name}.")

        for doc_batch, metadata_batch, id_batch in tqdm(
            zip(documents_batch, metadatas_batch, ids_batch, strict=True),
            total=n_batches,
        ):
            if self.embedder is None:
                raise ValueError("embedder is not set.")

            embedding_batch: EmbeddingResults = self.embedder.embed_text(doc_batch)

            n_retries = 0
            while n_retries < 3:
                try:
                    self._upsert(
                        client=client,
                        collection_name=self.collection_name,
                        ids=id_batch,
                        documents=doc_batch,
                        embeddings=embedding_batch.embeddings,
                        metadatas=metadata_batch,
                        **kwargs,
                    )
                    break
                except Exception as e:
                    self.logger.error(f"Error: {e}")
                    n_retries += 1
                    time.sleep(10)

                    if n_retries == 3:
                        raise e

    def _verify_metadata(self, metadatas: list[dict[str, str]]) -> None:
        """
        check if the key 'document' is included in metadatas
        """
        metadata_keys = [metadata for metadata in metadatas if "document" in metadata]
        if len(metadata_keys) > 0:
            raise ValueError("The key 'document' is reserved. Use another key.")


class BaseRemoteCollectionModule(BaseLocalCollectionModule, AbstractRemoteCollectionModule):  # type: ignore[misc]
    def __init__(
        self,
        url: str,
        collection_name: str,
        port: str,
        embedder: LLMModel | None = None,  # type: ignore
        logger: Any | None = None,
        batch_size: int = 100,
    ) -> None:
        self.url = url
        self.port = port
        self.embedder = embedder
        self.collection_name = collection_name
        self.logger = logger or default_logger
        self.batch_size = batch_size

    async def create_collection_async(self) -> None:
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        await self._create_collection_async(client=client, collection_name=self.collection_name)

    async def exists_async(self) -> bool:
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        return await self._exists_async(client=client, collection_name=self.collection_name)

    async def upsert_async(
        self,
        ids: list[str | int],
        documents: list[str],
        embeddings: list[list[float]],
        metadatas: list[dict[str, str]],
        **kwargs: Any,
    ) -> None:
        self._verify_metadata(metadatas)
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        await self._upsert_async(
            client=client,
            collection_name=self.collection_name,
            ids=ids,
            documents=documents,
            embeddings=embeddings,
            metadatas=metadatas,
            **kwargs,
        )

    async def delete_collection_async(self) -> None:
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        if await self._exists_async(client=client, collection_name=self.collection_name):
            await self._delete_collection_async(client=client, collection_name=self.collection_name)

    async def delete_record_async(self, **kwargs: Any) -> None:
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        await self._delete_record_async(
            client=client, collection_name=self.collection_name, **kwargs
        )

    async def run_async(
        self,
        documents: list[str],
        metadatas: Optional[list[dict[str, str]]] = None,
        **kwargs: Any,
    ) -> None:
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        if metadatas is not None and len(documents) != len(metadatas):
            raise ValueError(
                "The length of documents and metadatas must be the same. "
                f"Got {len(documents)} documents and {len(metadatas)} metadatas."
            )

        if metadatas is None:
            metadatas = [{} for _ in range(len(documents))]
        else:
            # check if the key 'collection' or 'document' is included in metadatas
            self._verify_metadata(metadatas)

        ids = [i for i in range(len(documents))]

        documents_batch = make_batch(documents, batch_size=self.batch_size)
        metadatas_batch = make_batch(metadatas, batch_size=self.batch_size)
        ids_batch = make_batch(ids, batch_size=self.batch_size)

        n_batches: int = math.ceil(len(documents) / self.batch_size)

        if not await self._exists_async(client=client, collection_name=self.collection_name):
            await self._create_collection_async(client=client, collection_name=self.collection_name)
            self.logger.info(f"Create collection {self.collection_name}.")

        for doc_batch, metadata_batch, id_batch in tqdm(
            zip(documents_batch, metadatas_batch, ids_batch, strict=True),
            total=n_batches,
        ):
            if self.embedder is None:
                raise ValueError("embedder is not set.")

            embedding_batch: EmbeddingResults = await self.embedder.embed_text_async(doc_batch)

            n_retries = 0
            while n_retries < 3:
                try:
                    await self._upsert_async(
                        client=client,
                        collection_name=self.collection_name,
                        ids=id_batch,
                        documents=doc_batch,
                        embeddings=embedding_batch.embeddings,
                        metadatas=metadata_batch,
                        **kwargs,
                    )
                    break
                except Exception as e:
                    self.logger.error(f"Error: {e}")
                    n_retries += 1
                    await asyncio.sleep(10)

                    if n_retries == 3:
                        raise e


class AbstractLocalRetrievalModule(ABC):
    @abstractmethod
    def get_client(self) -> Any:
        """
        return the async client object
        """
        raise NotImplementedError

    @abstractmethod
    def _retrieve(
        self,
        client: Any,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs: Any,
    ) -> RetrievalResults:
        raise NotImplementedError


class AbstractRemoteRetrievalModule(AbstractLocalRetrievalModule):
    @abstractmethod
    def get_client(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def get_async_client(self) -> Any:
        """
        return the async client object
        """
        raise NotImplementedError

    @abstractmethod
    def _retrieve(
        self,
        client: Any,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs: Any,
    ) -> RetrievalResults:
        raise NotImplementedError

    @abstractmethod
    async def _retrieve_async(
        self,
        client: Any,
        collection_name: str,
        query_vector: list[float],
        n_results: int,
        score_threshold: float,
        filter: Any | None = None,
        **kwargs: Any,
    ) -> RetrievalResults:
        raise NotImplementedError


class BaseLocalRetrievalModule(AbstractLocalRetrievalModule):
    def __init__(
        self,
        persistence_directory: Path | str,
        collection_name: str,
        embedder: LLMModel | None = None,  # type: ignore
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = False,
    ):
        self.persistence_directory = Path(persistence_directory)
        self.embedder = embedder
        self.collection_name = collection_name
        self.n_results = n_results
        self.score_threshold = score_threshold
        self.logger = logger or default_logger
        self.ascending = ascending

    def run(self, query: str, filter: Any | None = None, **kwargs: Any) -> RetrievalResults:
        client = self.get_client()

        if self.embedder is None:
            raise ValueError("embedder is not set.")

        embed: EmbeddingResults = self.embedder.embed_text(query)

        self.logger.info(f"Retrieve from collection {self.collection_name}...")
        retrieved: RetrievalResults = self._retrieve(
            client=client,
            collection_name=self.collection_name,
            query_vector=embed.embeddings[0],
            filter=filter,
            n_results=self.n_results,
            score_threshold=self.score_threshold,
            **kwargs,
        )

        retrieved.usage = embed.usage  # override

        return retrieved


class BaseRemoteRetrievalModule(BaseLocalRetrievalModule, AbstractRemoteRetrievalModule):
    def __init__(
        self,
        url: str,
        collection_name: str,
        port: str,
        embedder: LLMModel | None = None,  # type: ignore
        n_results: int = 4,
        score_threshold: float = 0.5,
        logger: Any | None = None,
        ascending: bool = False,
    ):
        self.url = url
        self.port = port
        self.embedder = embedder
        self.n_results = n_results
        self.score_threshold = score_threshold
        self.collection_name = collection_name
        self.ascending = ascending
        self.logger = logger or default_logger

    async def run_async(
        self, query: str, filter: Any | None = None, **kwargs: Any
    ) -> RetrievalResults:
        if inspect.iscoroutinefunction(self.get_async_client):
            client = await self.get_async_client()
        else:
            client = self.get_async_client()

        if self.embedder is None:
            raise ValueError("embedder is not set.")

        embed: EmbeddingResults = await self.embedder.embed_text_async(query)

        self.logger.info(f"Retrieve from collection {self.collection_name}...")
        retrieved: RetrievalResults = await self._retrieve_async(
            client=client,
            collection_name=self.collection_name,
            query_vector=embed.embeddings[0],
            filter=filter,
            n_results=self.n_results,
            score_threshold=self.score_threshold,
            **kwargs,
        )

        retrieved.usage = embed.usage  # override

        return retrieved


class BaseMetadataFilter(ABC):
    @abstractmethod
    def run(self, metadata: dict[str, str]) -> bool:
        raise NotImplementedError



import re
from pathlib import Path
from typing import Any

from pydantic import BaseModel, model_validator


class PromptTemplate(BaseModel):
    args: dict[str, Any] = {}
    template: str = ""

    @staticmethod
    def from_text_file(file_path: str | Path) -> "PromptTemplate":
        file_path = Path(file_path)

        with open(file_path, "r") as f:
            template = f.read()
        return PromptTemplate(template=template)

    def format(self) -> str:
        self._check_args(self.template, self.args)
        return self.template.format(**self.args)

    def set_args(self, **args: dict[str, Any]) -> "PromptTemplate":
        if self.template:
            self._check_args(self.template, args)
        self.args = args
        return self

    def set_template(self, template: str) -> "PromptTemplate":
        if self.args:
            self._check_args(template, self.args)
        self.template = template
        return self

    @model_validator(mode="after")
    def check_fields(self) -> "PromptTemplate":
        if self.args and self.template:
            self._check_args(self.template, self.args)
        return self

    @staticmethod
    def _check_args(
        template: str,
        args: dict[str, Any],
    ) -> None:
        def odd_repeat_pattern(s):
            return f"(?<!{s}){s}(?:{s}{s})*(?!{s})"

        pattern = re.compile(odd_repeat_pattern("{") + "[a-zA-Z0-9_]+" + odd_repeat_pattern("}"))
        pattern_sub = re.compile("{|}")
        found_args = set([pattern_sub.sub("", m) for m in pattern.findall(template)])
        input_args = set(args.keys())
        if found_args != input_args:
            if found_args:
                _found_args = "{" + ", ".join(found_args) + "}"
            else:
                _found_args = "{}"

            if input_args:
                _input_args = "{" + ", ".join(input_args) + "}"
            else:
                _input_args = "{}"

            raise ValueError(f"Template has args {_found_args} but {_input_args} were input")



import hashlib
import json
import logging
import os
import secrets
from typing import Any, cast

import boto3
from botocore.client import Config as BotoConfig
from botocore.exceptions import ClientError

from ..core.memory import BaseConversationMemory


class S3ConversationMemory(BaseConversationMemory):
    """
    A conversation memory that stores the conversation history in an S3 bucket.
    To use this memory, you need to create an S3 bucket in advance.

    Parameters
    ----------
    bucket : str
        The name of the S3 bucket to store the conversation history.
    object_key : str, optional
        The key of the object to store the conversation history. If not provided, a random key is generated.
    region_name : str, optional
        The region name of the S3 bucket, by default None.
    api_version : str, optional
        The API version of the S3 client, by default None.
    use_ssl : bool, optional
        Whether to use SSL for the S3 client, by default True.
    verify : str or bool, optional
        Whether to verify the SSL certificate for the S3 client, by default None.
    endpoint_url_env_name : str, optional
        The environment variable name for the S3 endpoint URL, by default None.
    aws_access_key_id_env_name : str, optional
        The environment variable name for the AWS access key ID, by default None.
    aws_secret_access_key_env_name : str, optional
        The environment variable name for the AWS secret access key, by default None.
    aws_session_token_env_name : str, optional
        The environment variable name for the AWS session token, by default None.
    boto_config : BotoConfig, optional
        The configuration for the S3 client, by default None.
    """

    def __init__(
        self,
        bucket: str,
        *,
        object_key: str | None = None,
        region_name: str | None = None,
        api_version: str | None = None,
        use_ssl: bool = True,
        verify: str | bool | None = None,
        endpoint_url_env_name: str | None = None,
        aws_access_key_id_env_name: str | None = None,
        aws_secret_access_key_env_name: str | None = None,
        aws_session_token_env_name: str | None = None,
        boto_config: BotoConfig | None = None,
    ) -> None:
        self.bucket = bucket
        self.object_key = object_key
        self.region_name = region_name
        self.api_version = api_version
        self.use_ssl = use_ssl
        self.verify = verify
        self.endpoint_url = os.getenv(endpoint_url_env_name) if endpoint_url_env_name else None
        self.aws_access_key_id = (
            os.getenv(aws_access_key_id_env_name) if aws_access_key_id_env_name else None
        )
        self.aws_secret_access_key = (
            os.getenv(aws_secret_access_key_env_name) if aws_secret_access_key_env_name else None
        )
        self.aws_session_token = (
            os.getenv(aws_session_token_env_name) if aws_session_token_env_name else None
        )
        self.boto_config = boto_config

        self.s3_client = boto3.client(
            "s3",
            region_name=self.region_name,
            api_version=self.api_version,
            use_ssl=self.use_ssl,
            verify=self.verify,
            endpoint_url=self.endpoint_url,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            config=self.boto_config,
        )

        try:
            # check at minimum
            self.s3_client.head_bucket(Bucket=self.bucket)
            logging.info(f"s3 connect success. bucket_name: {self.bucket}")
        except Exception as e:
            logging.error(f"s3 connect failed:{e}")
            raise

    def store(self, conversation_history: list[list[dict[str, Any]]]) -> None:
        """
        Store the conversation history in an S3 bucket.

        Parameters
        ----------
        conversation_history : list[list[dict[str, Any]]]
            The conversation history to store. The outer list represents the conversation turns,
            and the inner list represents the messages in each turn.
        """
        if self.object_key is None:
            self.object_key = hashlib.sha256(secrets.token_bytes(32)).hexdigest()
        try:
            json_data = json.dumps(conversation_history, ensure_ascii=False)
            self.s3_client.put_object(Bucket=self.bucket, Key=self.object_key, Body=json_data)
        except Exception as e:
            logging.error(f"s3 store failed: {e}")
            raise

    def load(self) -> list[list[dict[str, Any]]]:
        """
        Load the conversation history from an S3 bucket. If no history is found, return an empty list.
        The outer list represents the conversation turns, and the inner list represents the messages
        in each turn.

        Returns
        -------
        list[list[dict[str, Any]]]
            The conversation history. If no history is found, return an empty list.
        """
        if self.object_key is None:
            return []
        try:
            response = self.s3_client.get_object(Bucket=self.bucket, Key=self.object_key)
            content = response["Body"].read().decode("utf-8")
            conversation_history = json.loads(content)
            return cast(list[list[dict[str, Any]]], conversation_history)
        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "NoSuchKey":
                return []
            else:
                logging.error(f"s3 load failed: {e}")
                raise
        except Exception as e:
            logging.error(f"s3 load failedr: {e}")
            raise



from dataclasses import dataclass

from .error import (
    RetryLimitExceededError,
    TextResponseLimitExceededError,
    ToolCallLimitExceededError,
)


@dataclass
class AgentInternalContext:
    error_retries_count: int = 0
    max_error_retries: int = 3
    text_response_count: int = 0
    max_consecutive_text_response: int = 3
    tool_call_count: int = 0
    max_consecutive_tool_call: int = 3

    def __bool__(self) -> bool:
        return (
            self.error_retries_count < self.max_error_retries
            and self.text_response_count < self.max_consecutive_text_response
            and self.tool_call_count < self.max_consecutive_tool_call
        )

    def increment_error_retries_count(self) -> None:
        self.error_retries_count += 1

    def increment_repeat_text_response_count(self) -> None:
        self.text_response_count += 1

    def increment_repeat_tool_call_count(self) -> None:
        self.tool_call_count += 1

    def reset_error_retries_count(self) -> None:
        self.error_retries_count = 0

    def reset_repeat_text_response_count(self) -> None:
        self.text_response_count = 0

    def reset_repeat_tool_call_count(self) -> None:
        self.tool_call_count = 0

    def check(self) -> None:
        if self.error_retries_count >= self.max_error_retries:
            raise RetryLimitExceededError()
        elif self.tool_call_count >= self.max_consecutive_tool_call:
            raise ToolCallLimitExceededError()
        elif self.text_response_count >= self.max_consecutive_text_response:
            raise TextResponseLimitExceededError()



from typing import Any

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    class Config:
        arbitrary_types_allowed = True

    def update(self, **kwargs: Any) -> "BaseModel":
        return self.model_copy(update=kwargs)



import base64
import inspect
import io
import random
import string
from functools import partial
from inspect import signature
from typing import Any, Callable, Generator, Iterable
from urllib.parse import urlparse

import numpy as np
from PIL import Image
from pydantic import BaseModel


def get_variable_name_inspect(obj: object) -> str:
    if current_frame := inspect.currentframe():
        frame = current_frame.f_back
        while frame:
            for name, _obj in frame.f_globals.items():
                if _obj is obj:
                    return name
            frame = frame.f_back
    return ""


def make_batch(
    iterable: Iterable[Any], batch_size: int = 1, overlap: int = 0
) -> Generator[Iterable[Any], None, None]:
    if overlap >= batch_size:
        raise ValueError("overlap must be less than batch_size")

    length = len(iterable)

    st: int = 0
    while st < length:
        en = min(st + batch_size, length)
        batch = iterable[st:en]
        yield batch

        if en == length:
            break

        st += batch_size - overlap


def pil2bytes(image: Image.Image) -> bytes:
    num_byteio = io.BytesIO()
    image.save(num_byteio, format=image.format.lower() if image.format else "jpeg")
    num_byteio.seek(0)
    image_bytes = num_byteio.getvalue()
    return image_bytes


def image2bytes(image: Image.Image | np.ndarray | bytes) -> bytes:
    if isinstance(image, Image.Image):
        return pil2bytes(image)
    elif isinstance(image, np.ndarray):
        image_pil = Image.fromarray(image)
        return pil2bytes(image_pil)
    elif isinstance(image, bytes):
        return image
    else:
        raise ValueError(f"Type of {type(image)} is not supported for image.")


def base64_encode(data: Any):
    return base64.b64encode(data)


def base64_decode(data: Any):
    return base64.b64decode(data)


def utf8_to_bytes(text: str) -> bytes:
    return base64_decode(text.encode("utf-8"))


def encode_image(image: Image.Image | np.ndarray | bytes) -> str:
    image_bytes = image2bytes(image)
    encoded = base64_encode(image_bytes)
    return encoded.decode("utf-8")


def decode_image(image_encoded: str) -> Image.Image:
    image_encoded_utf = image_encoded.encode("utf-8")
    image_bytes = base64.b64decode(image_encoded_utf)
    byteio = io.BytesIO(image_bytes)
    return Image.open(byteio)


def generate_dummy_call_id(n: int) -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=n))


def model2func(model: BaseModel):
    def run_model(model: BaseModel, **kwargs):
        return model(**kwargs).run()

    return partial(run_model, model=model)


def is_valid_uri(text: str) -> bool:
    """Function to determine if a given string is a URI

    Args:
        text (str): The string to check

    Returns:
        bool: True if the string is a URI, False otherwise
    """
    try:
        result = urlparse(text)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def parse_arguments(
    obj: Callable[..., Any],
    include: set[str] | list[str] | None = None,
    exclude: set[str] | list[str] | None = None,
) -> set[str]:
    params = dict(signature(obj).parameters)

    if "self" in params:
        params.pop("self")

    if include is None:
        include = set()

    if exclude is None:
        exclude = set()

    if not isinstance(include, (set, list)):
        raise ValueError("include must be a set or list.")

    if not isinstance(exclude, (set, list)):
        raise ValueError("exclude must be a set or list.")

    if include:
        params = {k: v for k, v in params.items() if k in include}

    if exclude:
        params = {k: v for k, v in params.items() if k not in exclude}

    return set(params.keys())


def create_parameters(
    obj: Callable[..., Any],
    include: set[str] | list[str] | None = None,
    exclude: set[str] | list[str] | None = None,
    **kwargs: dict[str, Any],
) -> dict[str, Any]:
    params = parse_arguments(obj, include=include, exclude=exclude)
    invalid_params = set(kwargs.keys()) - params
    if invalid_params:
        raise ValueError(f"Invalid parameters: {invalid_params}, must be one of {params}")
    return {k: v for k, v in kwargs.items() if k in params}


def snake_to_camel(snake_str: str) -> str:
    if not snake_str:
        return ""

    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])



from abc import ABC, abstractmethod
from typing import Any


class BaseConversationMemory(ABC):
    @abstractmethod
    def store(self, conversation_history: list[list[dict[str, Any]]]) -> None:
        """
        Store the conversation history.

        Parameters
        ----------
        conversation_history : list[list[dict[str, Any]]]
            The conversation history to store. The outer list represents the conversation turns,
            and the inner list represents the messages in each turn.
        """
        raise NotImplementedError

    @abstractmethod
    def load(self) -> list[list[dict[str, Any]]]:
        """
        Load the conversation history. If no history is found, return an empty list.
        The outer list represents the conversation turns, and the inner list represents the messages
        in each turn.

        Returns
        -------
        list[list[dict[str, Any]]]
            The conversation history. If no history is found, return an empty list.
        """
        raise NotImplementedError



import inspect
from logging import Logger
from typing import Any, AsyncGenerator, Callable, Generator, Generic, Sequence

from .client import LLMClient
from .embedding import EmbeddingResults
from .logger import DEFAULT_LOGGER as default_logger
from .prompt import (
    AudioPrompt,
    ImagePrompt,
    Prompt,
    PromptType,
    SystemPrompt,
    TextPrompt,
    ToolCallPrompt,
)
from .response import (
    AudioResponse,
    ImageResponse,
    Response,
    ResponseType,
    TextResponse,
    ToolCallResponse,
)
from .tool import Tool
from .typing import ClientMessage, ClientMessageContent, ClientSystemMessage, ClientTool

LLMInput = (
    str
    | Prompt
    | PromptType
    | Response
    | ResponseType
    | list[str | Prompt | PromptType | ResponseType | Response]
)


class LLMModel(Generic[ClientMessage, ClientSystemMessage, ClientMessageContent, ClientTool]):
    """
    A high-level interface to interact with a language model client.

    Parameters
    ----------
    client : LLMClient
        Client to interact with.
    system_instruction : SystemPrompt, optional
        System instruction to generate text, by default None.
    tools : list[Callable[..., Any] | Tool], optional
        Tools that agent can use, by default None.
    logger : Logger, optional
        Logger to use, by default None.
    **kwargs : Any
        Additional arguments to pass to the client.
        What arguments are available depends on the client.
    """

    def __init__(
        self,
        client: LLMClient[ClientMessage, ClientSystemMessage, ClientMessageContent, ClientTool],
        system_instruction: SystemPrompt | None = None,
        tools: list[Callable[..., Any] | Tool] | None = None,
        logger: Logger | None = None,
        **kwargs: Any,
    ):
        self.client = client
        self.logger = logger or default_logger
        self.system_instruction = system_instruction
        self.tools = tools or []
        self.init_kwargs = kwargs

    def _prepare_client_request(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        tools: list[Callable[..., Any] | Tool] | None = None,
        **kwargs: Any,
    ) -> tuple[list[ClientMessage], ClientSystemMessage | None, Any]:
        messages = self._process_user_prompt(messages)

        all_kwargs: Any = {**self.init_kwargs, **kwargs}

        if __tools := tools or self.tools:
            _tools = self._prepare_tools(__tools)
            all_kwargs["tools"] = self.client.map_to_client_tools(tools=_tools)

        messages = self._convert_message_to_list(messages)
        mapped_messages = self._response_to_prompt(messages)

        self.logger.debug(f"Prompt: {mapped_messages[-1].contents}")

        prompt = self.client.map_to_client_prompts(mapped_messages)
        _system_instruction = self.client._setup_system_instruction(system_instruction)

        return prompt, _system_instruction, all_kwargs

    def generate_text(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        tools: list[Callable[..., Any] | Tool] | None = None,
        **kwargs: Any,
    ) -> Response:
        """
        Generate text based on the given prompt.

        Parameters
        ----------
        messages : LLMInput
            Prompt to generate text from.
        system_instruction : SystemPrompt | None, optional
            System instruction to generate text, by default None.
        tools : list[Callable[..., Any] | Tool] | None, optional
            Tools that agent can use, by default None.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        Response
            Generated text.
        """
        prompt, _system_instruction, all_kwargs = self._prepare_client_request(
            messages, system_instruction or self.system_instruction, tools, **kwargs
        )

        self.logger.info("{name}: Generating text".format(name=all_kwargs.get("name", "root")))
        response = self.client.generate_text(prompt, _system_instruction, **all_kwargs)

        self.logger.debug(f"Response: {response.contents}")

        return response

    async def generate_text_async(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        tools: list[Callable[..., Any] | Tool] | None = None,
        **kwargs: Any,
    ) -> Response:
        """
        Generate text based on the given prompt asynchronously.

        Parameters
        ----------
        messages : LLMInput
            Prompt to generate text from.
        system_instruction : SystemPrompt | None, optional
            System instruction to generate text, by default None.
        tools : list[Callable[..., Any] | Tool] | None, optional
            Tools that agent can use, by default None.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        Response
            Generated text.
        """
        prompt, _system_instruction, all_kwargs = self._prepare_client_request(
            messages, system_instruction or self.system_instruction, tools, **kwargs
        )

        self.logger.info("{name}: Generating text".format(name=all_kwargs.get("name", "root")))
        response = await self.client.generate_text_async(prompt, _system_instruction, **all_kwargs)

        self.logger.debug(f"Response: {response.contents}")

        return response

    def stream_text(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        tools: list[Callable[..., Any] | Tool] | None = None,
        **kwargs: Any,
    ) -> Generator[Response, None, None]:
        """
        Stream text based on the given prompt.

        Parameters
        ----------
        messages : LLMInput
            Prompt to generate text from.
        system_instruction : SystemPrompt | None, optional
            System instruction to generate text, by default None.
        tools : list[Callable[..., Any] | Tool] | None, optional
            Tools that agent can use, by default None.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Yields
        ----------
        Response
            Generated text.
        """
        prompt, _system_instruction, all_kwargs = self._prepare_client_request(
            messages, system_instruction or self.system_instruction, tools, **kwargs
        )

        self.logger.info("{name}: Generating text".format(name=all_kwargs.get("name", "root")))
        streamed_response = self.client.stream_text(prompt, _system_instruction, **all_kwargs)

        if inspect.isgenerator(streamed_response):
            for chunk in streamed_response:
                if chunk.is_last_chunk:
                    self.logger.debug(f"Response: {chunk.contents}")

                yield chunk
        else:
            raise ValueError(f"Expected a generator, but got {type(streamed_response)}")

    async def stream_text_async(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        tools: list[Callable[..., Any] | Tool] | None = None,
        **kwargs: Any,
    ) -> AsyncGenerator[Response, None]:
        """
        Stream text based on the given prompt asynchronously.

        Parameters
        ----------
        messages : LLMInput
            Prompt to generate text from.
        system_instruction : SystemPrompt | None, optional
            System instruction to generate text, by default None.
        tools : list[Callable[..., Any] | Tool] | None, optional
            Tools that agent can use, by default None.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Yields
        ----------
        Response
            Generated text.
        """
        prompt, _system_instruction, all_kwargs = self._prepare_client_request(
            messages, system_instruction or self.system_instruction, tools, **kwargs
        )

        self.logger.info("{name}: Generating text".format(name=all_kwargs.get("name", "root")))
        streamed_response = self.client.stream_text_async(prompt, _system_instruction, **all_kwargs)

        if inspect.isasyncgen(streamed_response):
            async for chunk in streamed_response:
                if chunk.is_last_chunk:
                    self.logger.debug(f"Response: {chunk.contents}")

                yield chunk
        else:
            raise ValueError(f"Expected a async generator, but got {type(streamed_response)}")

    def generate_image(self, prompt: str, **kwargs: Any) -> Response:
        """
        Generate an image based on the given prompt.

        Parameters
        ----------
        prompt : str
            Prompt to generate the image from.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        Response
            Generated image.
        """
        all_kwargs: Any = {**self.init_kwargs, **kwargs}

        self.logger.info("Generating image")
        response = self.client.generate_image(prompt, **all_kwargs)

        return response

    async def generate_image_async(self, prompt: str, **kwargs: Any) -> Response:
        """
        Generate an image based on the given prompt asynchronously.

        Parameters
        ----------
        prompt : str
            Prompt to generate the image from.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        Response
            Generated image.
        """
        all_kwargs: Any = {**self.init_kwargs, **kwargs}

        self.logger.info("Generating image")
        response = await self.client.generate_image_async(prompt, **all_kwargs)

        return response

    def generate_video(self, prompt: str, **kwargs: Any) -> Response:
        raise NotImplementedError

    async def generate_video_async(self, prompt: str, **kwargs: Any) -> Response:
        raise NotImplementedError

    def generate_audio(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        **kwargs: Any,
    ) -> Response:
        """
        Generate audio based on the given prompt.

        Parameters
        ----------
        messages : LLMInput
            Prompt to generate audio from.
        system_instruction : SystemPrompt | None, optional
            System instruction to generate audio, by default None.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        Response
            Generated audio.
        """
        prompt, _system_instruction, all_kwargs = self._prepare_client_request(
            messages, system_instruction, None, **kwargs
        )

        self.logger.info("Generating audio")
        response = self.client.generate_audio(prompt, _system_instruction, **all_kwargs)

        return response

    async def generate_audio_async(
        self,
        messages: LLMInput,
        system_instruction: SystemPrompt | None = None,
        **kwargs: Any,
    ) -> Response:
        """
        Generate audio based on the given prompt asynchronously.

        Parameters
        ----------
        messages : LLMInput
            Prompt to generate audio from.
        system_instruction : SystemPrompt | None, optional
            System instruction to generate audio, by default None.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        Response
            Generated audio.
        """
        prompt, _system_instruction, all_kwargs = self._prepare_client_request(
            messages, system_instruction, None, **kwargs
        )

        self.logger.info("Generating audio")
        response = await self.client.generate_audio_async(prompt, _system_instruction, **all_kwargs)

        return response

    def embed_text(self, texts: Sequence[str], **kwargs: Any) -> EmbeddingResults:
        """
        Embed text into a vector representation.

        Parameters
        ----------
        texts : Sequence[str]
            Texts to embed.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        EmbeddingResults
            Embedding results.
        """
        all_kwargs: Any = {**self.init_kwargs, **kwargs}

        self.logger.info("Embedding text")
        return self.client.embed_text(texts, **all_kwargs)

    async def embed_text_async(self, texts: Sequence[str], **kwargs: Any) -> EmbeddingResults:
        """
        Embed text into a vector representation asynchronously.

        Parameters
        ----------
        texts : Sequence[str]
            Texts to embed.
        **kwargs : Any
            Additional arguments to pass to the client.
            What arguments are available depends on the client.

        Returns
        ----------
        EmbeddingResults
            Embedding results.
        """
        all_kwargs: Any = {**self.init_kwargs, **kwargs}

        self.logger.info("Embedding text")
        return await self.client.embed_text_async(texts, **all_kwargs)

    def _convert_message_to_list(
        self, message: list[Prompt | Response] | Prompt | Response
    ) -> list[Prompt | Response]:
        if not isinstance(message, list):
            return [message]

        return message

    def _response_to_prompt(self, msg_or_res: list[Prompt | Response]) -> list[Prompt]:
        """
        Convert a list of Response objects to a list of Prompt objects.
        Prompt objects are used as it is.
        """
        messages: list[Prompt] = []
        for _msg_or_res in msg_or_res:
            if isinstance(_msg_or_res, Prompt):
                messages.append(_msg_or_res)
                continue

            contents: list[PromptType] = []

            for content in _msg_or_res.contents or []:
                if isinstance(content, TextResponse):
                    contents.append(TextPrompt(text=content.text))
                elif isinstance(content, ImageResponse):
                    contents.append(
                        ImagePrompt(image=content.image, format=content.format or "jpeg")
                    )
                elif isinstance(content, AudioResponse):
                    contents.append(
                        AudioPrompt(
                            audio=content.audio,
                            mime_type=content.mime_type,
                            audio_id=content.audio_id,
                        )
                    )

                elif isinstance(content, ToolCallResponse):
                    contents.append(
                        ToolCallPrompt(
                            call_id=content.call_id,
                            name=content.name or "",
                            args=content.args or "{}",
                        )
                    )
                elif isinstance(content, PromptType) and content:
                    contents.append(content)
                else:
                    raise ValueError(f"Invalid content type: {type(content)}")

            messages.append(
                Prompt(
                    contents=contents,
                    role=_msg_or_res.role,
                    name=_msg_or_res.name,
                )
            )

        return messages

    def _prepare_tools(self, tools: list[Callable[..., Any] | Tool]) -> list[Tool]:
        outputs: list[Tool] = []
        for tool in tools:
            if isinstance(tool, Tool):
                outputs.append(tool)
            elif callable(tool):
                # self.logger.debug(f"Preparing tools: {tool.__name__}")
                outputs.append(Tool(tool=tool))
            else:
                raise ValueError(f"Invalid tool type: {type(tool)}")
        return outputs

    def _process_user_prompt(self, prompt: LLMInput) -> list[Prompt | Response]:
        if isinstance(prompt, (Prompt, Response)):
            return [prompt]
        elif isinstance(prompt, (str, PromptType)):
            return [Prompt(role="user", contents=prompt)]
        elif isinstance(prompt, ResponseType):
            return [Response(role="assistant", contents=[prompt])]
        elif isinstance(prompt, list):
            include_prompt_or_response = False
            include_content = False

            for p in prompt:
                if isinstance(p, (Prompt, Response)):
                    include_prompt_or_response = True
                elif isinstance(p, (str, PromptType)):
                    include_content = True

            if include_prompt_or_response and include_content:
                raise ValueError(
                    "Prompt types or roles are ambiguous. Don't mix Prompt/Response and str/content."
                )

            if include_prompt_or_response:
                messages = prompt
            elif include_content:
                messages = [Prompt(role="user", contents=prompt)]
            else:
                raise ValueError("Invalid prompt type")

            return messages
        else:
            raise ValueError("Invalid prompt type. Please provide the correct prompt type.")



from typing import Generic, TypeVar

from .prompt import Prompt
from .pydantic import BaseModel
from .response import Response

MessageType = TypeVar("MessageType", Prompt, Response)


class Message(BaseModel, Generic[MessageType]):
    message: MessageType



import json
import os
from typing import Any, cast

from ..core.memory import BaseConversationMemory


class JSONConversationMemory(BaseConversationMemory):
    """
    A conversation memory that stores the conversation history in a local JSON file.

    Parameters
    ----------
    path : str, optional
        The path to the JSON file to store the conversation history, by default "conversation_memory.json"
    exist_ok : bool, optional
        If True, do not raise an error if the file already exists and reuses it.
        If False, delete the file if it exists and create a new one, by default False.
    """

    def __init__(self, path: str = "conversation_memory.json", exist_ok: bool = False) -> None:
        self.path = path

        if not exist_ok and os.path.exists(self.path):
            os.remove(self.path)

    def store(self, conversation_history: list[list[dict[str, Any]]]) -> None:
        """
        Store the conversation history in a local JSON file.

        Parameters
        ----------
        conversation_history : list[list[dict[str, Any]]]
            The conversation history to store. The outer list represents the conversation turns,
            and the inner list represents the messages in each turn.
        """
        with open(self.path, "w") as f:
            json.dump(conversation_history, f, ensure_ascii=False, indent=2)

    def load(self) -> list[list[dict[str, Any]]]:
        """
        Load the conversation history from a local JSON file. If no history is found, return an empty list.
        The outer list represents the conversation turns, and the inner list represents the messages
        in each turn.

        Returns
        -------
        list[list[dict[str, Any]]]
            The conversation history. If no history is found, return an empty list.
        """
        if os.path.exists(self.path):
            with open(self.path, "r") as f:
                conversation_history = json.load(f)
            return cast(list[list[dict[str, Any]]], conversation_history)
        else:
            return []
